import type { Completion } from '@codemirror/autocomplete';
import type { CompletionContext } from '@codemirror/autocomplete';
import { parser as dentakuParser } from './dentaku.grammar';
import { LanguageSupport } from '@codemirror/language';
import { linter } from '@codemirror/lint';
import type { LintSource } from '@codemirror/lint';
import { LRLanguage } from '@codemirror/language';
import { Options } from 'tsup';

/** Arity config for built-in and custom functions  */
declare type Arity = {
    /** @default 0 */
    minArgs?: number;
    /** @default Infinity */
    maxArgs?: number;
};
export { Arity }
export { Arity as Arity_alias_1 }

/**
 * The supported built-in functions names and their config.
 *
 * When `null` config is provided, the `minArgs` is assumed to be 0 and `maxArg`
 * to be Infinity
 */
export declare const builtInFunctions: {
    acos: {
        minArgs: number;
        maxArgs: number;
    };
    acosh: {
        minArgs: number;
        maxArgs: number;
    };
    asin: {
        minArgs: number;
        maxArgs: number;
    };
    asinh: {
        minArgs: number;
        maxArgs: number;
    };
    atan: {
        minArgs: number;
        maxArgs: number;
    };
    atan2: {
        minArgs: number;
        maxArgs: number;
    };
    atanh: {
        minArgs: number;
        maxArgs: number;
    };
    cbrt: {
        minArgs: number;
        maxArgs: number;
    };
    cos: {
        minArgs: number;
        maxArgs: number;
    };
    cosh: {
        minArgs: number;
        maxArgs: number;
    };
    erf: {
        minArgs: number;
        maxArgs: number;
    };
    erfc: {
        minArgs: number;
        maxArgs: number;
    };
    exp: {
        minArgs: number;
        maxArgs: number;
    };
    frexp: {
        minArgs: number;
        maxArgs: number;
    };
    gamma: {
        minArgs: number;
        maxArgs: number;
    };
    hypot: {
        minArgs: number;
        maxArgs: number;
    };
    ldexp: {
        minArgs: number;
        maxArgs: number;
    };
    lgamma: {
        minArgs: number;
        maxArgs: number;
    };
    log: {
        minArgs: number;
        maxArgs: number;
    };
    log10: {
        minArgs: number;
        maxArgs: number;
    };
    log2: {
        minArgs: number;
        maxArgs: number;
    };
    sin: {
        minArgs: number;
        maxArgs: number;
    };
    sinh: {
        minArgs: number;
        maxArgs: number;
    };
    sqrt: {
        minArgs: number;
        maxArgs: number;
    };
    tan: {
        minArgs: number;
        maxArgs: number;
    };
    tanh: {
        minArgs: number;
        maxArgs: number;
    };
    if: {
        minArgs: number;
        maxArgs: number;
    };
    and: {
        minArgs: number;
        maxArgs: number;
    };
    or: {
        minArgs: number;
        maxArgs: number;
    };
    xor: {
        minArgs: number;
        maxArgs: number;
    };
    not: {
        minArgs: number;
        maxArgs: number;
    };
    switch: {
        minArgs: number;
        maxArgs: number;
    };
    min: {
        minArgs: number;
        maxArgs: number;
    };
    max: {
        minArgs: number;
        maxArgs: number;
    };
    sum: {
        minArgs: number;
        maxArgs: number;
    };
    avg: {
        minArgs: number;
        maxArgs: number;
    };
    count: {};
    round: {
        minArgs: number;
        maxArgs: number;
    };
    roundup: {
        minArgs: number;
        maxArgs: number;
    };
    rounddown: {
        minArgs: number;
        maxArgs: number;
    };
    abs: {
        minArgs: number;
        maxArgs: number;
    };
    case: {
        minArgs: number;
        maxArgs: number;
    };
    left: {
        minArgs: number;
        maxArgs: number;
    };
    right: {
        minArgs: number;
        maxArgs: number;
    };
    mid: {
        minArgs: number;
        maxArgs: number;
    };
    len: {
        minArgs: number;
        maxArgs: number;
    };
    substitute: {
        minArgs: number;
        maxArgs: number;
    };
    concat: {
        minArgs: number;
        maxArgs: number;
    };
    contains: {
        minArgs: number;
        maxArgs: number;
    };
    map: {
        minArgs: number;
        maxArgs: number;
    };
    filter: {
        minArgs: number;
        maxArgs: number;
    };
    all: {
        minArgs: number;
        maxArgs: number;
    };
    any: {
        minArgs: number;
        maxArgs: number;
    };
    pluck: {
        minArgs: number;
        maxArgs: number;
    };
    duration: {
        minArgs: number;
        maxArgs: number;
    };
};

export declare type BuiltInFunctionsType = keyof typeof builtInFunctions;

export declare const builtInOperators: readonly ["+", "-", "*", "/", "%", "^", "<", ">", "<=", ">=", "=", "==", "!=", "<>", "&", "|", ">>", "<<"];

export declare const default_alias: Options | Options[] | ((overrideOptions: Options) => Options | Options[] | Promise<Options | Options[]>);

/**
 * The full Dentaku language support, including autocompletion and linting.
 *
 * @example
 * const codeMirrorExtensions = [
 *   ...otherExtensions,
 *   dentaku(),
 * ];
 */
export declare function dentaku({ linterOptions: { codeMirrorConfig, ...linterOptions }, completionOptions, }?: {
    linterOptions?: {
        codeMirrorConfig?: Parameters<typeof linter>[1];
    } & DentakuLinterOptions;
    completionOptions?: DentakuLanguageCompletionOptions;
}): LanguageSupport;

declare function dentakuCompletions({ variableEntries, customFunctionEntries, makeEntryForBuiltInFunctions, makeEntryForBuiltInOperators, }?: DentakuLanguageCompletionOptions): (context: CompletionContext) => {
    options: Completion[];
    from: number;
    validFor?: undefined;
} | {
    options: {
        label: string;
        type: string;
    }[];
    from: number;
    validFor: RegExp;
} | {
    options: Completion[];
    from: number;
    validFor: RegExp;
} | undefined;
export { dentakuCompletions }
export { dentakuCompletions as dentakuCompletions_alias_1 }

/**
 * Only the parser and syntax highlighter for Dentaku.
 *
 * Use this only if you want to omit/overwrite either linting or autocompletion.
 */
export declare const dentakuLanguage: LRLanguage;

declare interface DentakuLanguageCompletionOptions {
    /**
     * CodeMirror `Completion` objects for known variables.
     *
     * @example
     * ["a", "b"].map(name => ({ label: name, type: "variable" }))
     */
    variableEntries?: Array<Completion>;
    /**
     * CodeMirror `Completion` objects for custom functions.
     *
     * @example
     * ["fnA", "fnB"].map(name => ({ label: name, type: "function" }))
     */
    customFunctionEntries?: Array<Completion>;
    /**
     * Converter of built-in Dentaku functions into `Completion` objects.
     * Allows you to add additional information to completions like descriptions.
     */
    makeEntryForBuiltInFunctions?: (name: BuiltInFunctionsType) => Completion | null | false;
    /**
     * Converter of built-in Dentaku operators into `Completion` objects.
     * Allows you to add additional information to completions like descriptions.
     */
    makeEntryForBuiltInOperators?: (name: (typeof builtInOperators)[number]) => Completion | null | false;
}
export { DentakuLanguageCompletionOptions }
export { DentakuLanguageCompletionOptions as DentakuLanguageCompletionOptions_alias_1 }

declare function dentakuLinter({ knownVariables, customFunctions, messages, }?: DentakuLinterOptions): LintSource;
export { dentakuLinter }
export { dentakuLinter as dentakuLinter_alias_1 }

declare interface DentakuLinterOptions {
    /** List of variable names to not treat as undefined. */
    knownVariables?: Array<string>;
    /** Custom functions. */
    customFunctions?: Record<string, Arity>;
    /** Custom error messages. */
    messages?: ErrorMessages;
}
export { DentakuLinterOptions }
export { DentakuLinterOptions as DentakuLinterOptions_alias_1 }

export { dentakuParser }

/** Message codes that can be overridden with custom messages. */
declare interface ErrorMessages {
    /** Generic "invalid syntax" message when no better message can be inferred. */
    invalidSyntax: string;
    /** Missing a closing bracket in a list of function arguments. */
    closingBracketMissing: string;
    /** Expected a comma before this function argument. */
    expectedCommaBefore: string;
    /** Expected an operator before this expression. */
    expectedOperatorBefore: string;
    /** The function name needs parentheses to be called. */
    callParenthesesMissing: string;
    /** This variable is not defined. */
    undefinedVariable: string;
    /** Missing child property. */
    missingChildProperty: string;
    /** Expected exactly that many parameters, found this many. */
    expectedExactArgumentCount: (actualCount: number, expectedCount: number) => string;
    /** Expected at most that many parameters, found this many. */
    expectedMaximumArgumentCount: (actualCount: number, maxCount: number) => string;
    /** Expected at least that many parameters, found this many. */
    expectedMinimumArgumentCount: (actualCount: number, minCount: number) => string;
    /** Expected between that and that many parameters, found this many. */
    expectedArgumentCountRange: (actualCount: number, minCount: number, maxCount: number) => string;
}
export { ErrorMessages }
export { ErrorMessages as ErrorMessages_alias_1 }

/** Special character reserved for invalid syntax trees. */
declare const invalidSyntax = "\u26A0";
export { invalidSyntax }
export { invalidSyntax as invalidSyntax_alias_1 }

export { }
